/*
 * area.cc
 * 
 * Electrostatic PDE solver element area implementation. 
 * 
 * Copyright (c) 2004 by Wolfgang Wieser (wwieser@gmx.de) 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <sys/time.h>
#include <assert.h>

#include "elscore.h"


long msec_elapsed(const struct timeval *old,const struct timeval *current)
{
	struct timeval _curr;
	if(!current)
	{
		gettimeofday(&_curr,NULL);
		current=&_curr;
	}
	return(
		(current->tv_sec  - old->tv_sec )*1000L + 
		(current->tv_usec - old->tv_usec)/1000L );
}


int EStatArea::RunSolver(SolverParams *sp)
{
	if(width<1 || height<1)
	{  return(1);  }
	
	struct timeval start_time;
	if(sp->timeslice>=0)
	{  gettimeofday(&start_time,NULL);  }
	
	for(int i=0; i<16; i++)
	{
		// Sweep increment: -1 or +1. 
		int sinc;
		// Sweep direction: 0=row-by-row, 1=column-by-column. 
		int sdir;
		switch(sp->shuffle)
		{
			case SolverParams::NoShuffle:  sinc=+1;  sdir=0;  break;
			case SolverParams::XpXmYpYm:
				switch(sp->iter%4)  // sweep direction
				{
					case 0:  sinc=+1;  sdir=0;  break;
					case 1:  sinc=-1;  sdir=0;  break;
					case 2:  sinc=+1;  sdir=1;  break;
					case 3:  sinc=-1;  sdir=1;  break;
				}
				break;
			case SolverParams::XpYpXmYm:
				switch(sp->iter%4)  // sweep direction
				{
					case 0:  sinc=+1;  sdir=0;  break;
					case 1:  sinc=+1;  sdir=1;  break;
					case 2:  sinc=-1;  sdir=0;  break;
					case 3:  sinc=-1;  sdir=1;  break;
				}
				break;
			default: assert(9);
		}
		
		double diff_max;
		switch(sp->method)
		{
			case SolverParams::ThomasTDMA:
				diff_max=TDMA_Iterate(sp,sdir,sinc);
				break;
			case SolverParams::GaussSeidel:
				diff_max=GaussSeidel_Iterate(sp,sdir,sinc);
				break;
			default: assert(0);
		}
		++sp->iter;
		if(diff_max<sp->eps)
		{  return(1);  }
		if(sp->maxiter && sp->iter>=sp->maxiter)
		{  return(-1);  }
		
		if(sp->timeslice>=0)
		{
			long msec=msec_elapsed(&start_time,NULL);
			if(msec>sp->timeslice)  break;
		}
	}
	
	return(0);
}


int EStatArea::SetupArea(AreaSetupPars *asp)
{
	_CleanupArea();
	
	width= asp->width<1  ? 1 : asp->width;
	height=asp->height<1 ? 1 : asp->height;
	
	EElement::Type back=asp->background;
	
	// Set up element array: 
	size_t n_elem=size_t(width)*size_t(height);
	_elem=(EElement*)malloc(n_elem*sizeof(EElement));
	assert(_elem);
	
	// Fill area: 
	Reset(asp->background);
	
	return(0);
}


void EStatArea::GetAreaSetup(AreaSetupPars *asp)
{
	asp->width=width;
	asp->height=height;
	asp->background=EElement::_LAST;  // (= do not change)
	asp->valid=1;
}


int EStatArea::UpdateElement(int cx,int cy,double phi,EElement::Type type)
{
	if(cx<0 || cy<0 || cx>=width || cy>=height)  return(0);
	
	int chg=0;
	
	EElement &e=elem(cx,cy);
	if(e.phi!=phi)
	{  e.phi=phi;  chg=1;  }
	if(e.type!=type)
	{
		--nelem[e.type];
		e.type=type;  chg=1;
		++nelem[e.type];
	}
	
	if(minphi>e.phi)
	{  minphi=e.phi;  chg=2;  }
	if(maxphi<e.phi)
	{  maxphi=e.phi;  chg=2;  }
	
	if(chg)
	{  chunks_dirty=1;  gs_sys_dirty=1;  }
	
	return(chg);
}


void EStatArea::Reset(EElement::Type back)
{
	_ResetCounters();
	
	if(back<0 || back>=EElement::_LAST)
		back=EElement::Floating;
	
	size_t n_elem=size_t(width)*size_t(height);
	// Fill elements: 
	for(EElement *i=_elem,*cae=i+n_elem; i<cae; i++)
	{
		i->type=back;
		i->phi=0.0;
		++nelem[i->type];
	}
	
	// Set anything for max/min since all potentials are 0 currently. 
	minphi=0.0;
	maxphi=0.0;
	
	chunks_dirty=1;
	gs_sys_dirty=1;
}


void EStatArea::SetAllFloating(double phi)
{
	for(int y=0; y<height; y++)
	{
		for(int x=0; x<width; x++)
		{
			EElement &e=elem(x,y);
			if(e.type==EElement::Floating)
			{  e.phi=phi;  }
		}
	}
	
	if(minphi>phi)
	{  minphi=phi;  }
	if(maxphi<phi)
	{  maxphi=phi;  }
}


void EStatArea::_ResetCounters()
{
	for(int i=0; i<EElement::_LAST; i++)
	{  nelem[i]=0;  }
}


void EStatArea::_CleanupArea()
{
	FreeChunkLines(ck_rows,height);
	FreeChunkLines(ck_cols,width);
	if(ck_rows)  {  free(ck_rows);  ck_rows=NULL;  }
	if(ck_cols)  {  free(ck_cols);  ck_cols=NULL;  }
	chunks_dirty=1;  // Correct. 
	
	FreeGSSys();
	gs_sys_dirty=1;
	
	if(_elem)   free(_elem);   _elem=NULL;
	minphi=0.0;
	maxphi=0.0;
	_ResetCounters();
}


EStatArea::EStatArea()
{
	width=0;
	height=0;
	_elem=NULL;
	_ResetCounters();
	
	ck_rows=NULL;
	ck_cols=NULL;
	chunks_dirty=1;  // Correct. 
	
	gs_n=0;
	gs_rows=NULL;
	gs_cols=NULL;
	gs_coeff_buf=NULL;
	gs_sys_dirty=1;
	
	#if 0
	AreaSetupPars asp;
	asp.width=200;
	asp.height=200;
	asp.background=EElement::Floating;
	asp.valid=1;
	SetupArea(&asp);
	#endif
}

EStatArea::~EStatArea()
{
	_CleanupArea();
}
