/*
 * elscore.h
 * 
 * Electrostatic PDE solver core header file. 
 * 
 * Copyright (c) 2004 by Wolfgang Wieser (wwieser@gmx.de) 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#ifndef _ELSTAT_ELSCORE_H_
#define _ELSTAT_ELSCORE_H_ 1

#include <stdlib.h>
#include "gauss-seidel.h"

// See tridiag.cc: 
extern void SolveTridiag_M1_TDMA(int n,double *d,double *y,double *x);


// This is an elementary cell: 
struct EElement
{
	// The potential value of this element: 
	double phi;
	
	// The type of this element: 
	enum Type
	{
		Floating=0,  // keep that zero
		Fixed,
		_LAST
	};
	Type type;
}/*__attribute__((packed))*/;


// As to be passed to EStatArea::SetupArea();
struct AreaSetupPars
{
	int width,height;
	EElement::Type background;
	
	bool valid;
	AreaSetupPars()  {  valid=false;  }
	~AreaSetupPars()  { }
};


// Parameters for the PDE solver. 
struct SolverParams
{
	enum Method
	{
		GaussSeidel,
		ThomasTDMA,
	} method;
	enum ShuffleMode
	{
		NoShuffle,
		XpXmYpYm,
		XpYpXmYm,
	} shuffle;
	int maxiter;   // max number of iterations
	double eps;    // convergation tolerance
	double omega;  // 1..2 for overrelaxation, 1.0 for normal method
	int timeslice; // time slice in msec
	bool periodic; // periodic BC?
	
	int iter;  // Iteration counter; initialize to 0. 
	
	bool valid;
	SolverParams() { valid=0; iter=0; }
};


// The complete simulation area consisting of elements: 
class EStatArea
{
	public:
		struct LChunk
		{
			int a,b;
		};

		struct ChunkLine
		{
			int nchunks;  // -1 -> all floating; -2 -> all fixed
			LChunk *chunk;
		};

	private:
		// Size of the area: 
		int width,height;
		
		// Array of elements; size width*height. Use elem() to access. 
		EElement *_elem;
		inline EElement &elem(int x,int y)
			{  return(_elem[y*size_t(width)+x]);  }
		
		// Counters: count how many elements are of specified type: 
		size_t nelem[EElement::_LAST];
		
		// Pointers to the area elements with the max/min potential value: 
		double maxphi,minphi;
		
		// Whole area in chunk lines (for TDMA): 
		ChunkLine *ck_rows,*ck_cols;
		bool chunks_dirty;  // Need to re-generate the chunk lines?
		
		// Whole linar system infomation (for Gauss-Seidel): 
		int gs_n;  // array size
		GaussSeidelRow_M1<double> *gs_rows,*gs_cols;
		char *gs_coeff_buf;
		bool gs_sys_dirty;
		bool gs_sys_periodic;  // GS system was made for periodic BC?
		
		// Cleanup...
		void _CleanupArea();
		
		// Reset the nelem[] counters: 
		void _ResetCounters();
		
		// Compute the chunk lines. 
		void ComputeChunkLines();
		void FreeChunkLines(ChunkLine *cl_array,int size);
		
		// Setup/cleanup the Gauss-Seidel system: 
		inline void _GSSetupElement_AddNeigh(int x,int y,
			GaussSeidelRow_M1<double> *rp);
		inline void _GSSetupElement_AddNeigh_B(int x,int y,
			GaussSeidelRow_M1<double> *rp);
		void _GSSetupElement(int x,int y,GaussSeidelRow_M1<double> *rp,
			bool periodic_bc);
		void SetupGSSys(SolverParams *sp);
		void FreeGSSys();
		
		// Solver, internal: 
		double TDMA_Iterate(SolverParams *sp,int sdir,int sinc);
		double _SolveTDM_H(double *diag,double *phi,
			int a,int b,int sy,int syu,int syl,double omega,double omega1);
		double _SolveTDM_V(double *diag,double *phi,
			int a,int b,int sx,int sxu,int sxl,double omega,double omega1);
		double GaussSeidel_Iterate(SolverParams *sp,int sdir,int sinc);
	public:
		EStatArea();
		~EStatArea();
		
		// Query size: 
		int Width()   {  return(width);  }
		int Height()  {  return(height);  }
		
		// Get min/max potential element: 
		double MinPhi()  {  return(minphi);  }
		double MaxPhi()  {  return(maxphi);  }
		
		// Get element reference: 
		const EElement &Elem(int x,int y)
			{  return(elem(x,y));  }
		
		// Update element. Use this instead of modifying directly. 
		// Returns 2, if the minp/maxp values changed, 1 if elem changed 
		// and 0 if nothing changed. 
		int UpdateElement(int cx,int cy,double phi,EElement::Type type);
		
		// Fill complete ara with specified background: 
		void Reset(EElement::Type back);
		
		// Set all floating elemnts to passed value. 
		void SetAllFloating(double phi);
		
		// Change size etc.; will clear the complete area. 
		// Returns 0 on success. 
		int SetupArea(AreaSetupPars *asp);
		
		// Get current parameter setup: 
		void GetAreaSetup(AreaSetupPars *asp);
		
		// Run the PDE solver: 
		// Return value: 
		//  1 -> OK, converged
		// -1 -> failed
		//  0 -> returned but call again (time slice over)
		int RunSolver(SolverParams *sp);
};

#endif  /* _ELSTAT_ELSCORE_H_ */
