/*
 * solve-tdma.cc
 * 
 * Electrostatic PDE solver Gauss-Seidel solver implementation. 
 * 
 * Copyright (c) 2004 by Wolfgang Wieser (wwieser@gmx.de) 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#include <stdio.h>
#include <math.h>
#include <assert.h>

#include "elscore.h"


double EStatArea::GaussSeidel_Iterate(SolverParams *sp,
	int sdir,int sinc)
{
	assert(width>=2 && height>=2);
	
	if(gs_sys_periodic!=sp->periodic)
	{  gs_sys_dirty=1;  }
	if(gs_sys_dirty)
	{  SetupGSSys(sp);  }
	
	assert(gs_n==nelem[EElement::Floating]);
	
	double diffmax=GaussSeidelIterate_M1(gs_n,
		sdir ? gs_cols : gs_rows,sp->eps,
		/*maxiter=*/sinc/*correct!*/,sp->omega);
	
	return(diffmax);
}


inline void EStatArea::_GSSetupElement_AddNeigh(int x,int y,
	GaussSeidelRow_M1<double> *rp)
{
	EElement &e=elem(x,y);
	if(e.type==EElement::Floating)
	{  rp->x[rp->nent++]=&e.phi;  }
	else
	{  rp->y+=e.phi;  }
}

inline void EStatArea::_GSSetupElement_AddNeigh_B(int x,int y,
	GaussSeidelRow_M1<double> *rp)
{
	// Border element with check. 
	if(x<0 || x==width || y<0 || y==height)
	{  rp->A0-=1.0;  }
	else
	{  _GSSetupElement_AddNeigh(x,y,rp);  }
}

void EStatArea::_GSSetupElement(int x,int y,GaussSeidelRow_M1<double> *rp,
	bool periodic_bc)
{
	rp->nent=1;
	rp->A0=4.0;
	rp->x[0]=&elem(x,y).phi;
	rp->y=0.0;
	
	if(x && y && x!=width-1 && y!=height-1)
	{
		// Central element: 
		_GSSetupElement_AddNeigh(x,y-1,rp);
		_GSSetupElement_AddNeigh(x-1,y,rp);
		_GSSetupElement_AddNeigh(x+1,y,rp);
		_GSSetupElement_AddNeigh(x,y+1,rp);
	}
	else if(periodic_bc)
	{
		// Border element. 
		_GSSetupElement_AddNeigh(x,(y ? y : height)-1,rp);
		_GSSetupElement_AddNeigh((x ? x : width)-1,y,rp);
		int tmp=x+1;  if(tmp==width) tmp=0;
		_GSSetupElement_AddNeigh(tmp,y,rp);
		tmp=y+1;  if(tmp==height) tmp=0;
		_GSSetupElement_AddNeigh(x,tmp,rp);
	}
	else
	{
		// Border element. 
		_GSSetupElement_AddNeigh_B(x,y-1,rp);
		_GSSetupElement_AddNeigh_B(x-1,y,rp);
		_GSSetupElement_AddNeigh_B(x+1,y,rp);
		_GSSetupElement_AddNeigh_B(x,y+1,rp);
	}
}


void EStatArea::SetupGSSys(SolverParams *sp)
{
	fprintf(stderr,"(SGS)");
	
	FreeGSSys();
	
	gs_n=nelem[EElement::Floating];
	assert(gs_n>=0);
	gs_rows=(GaussSeidelRow_M1<double>*)malloc(
		gs_n*sizeof(GaussSeidelRow_M1<double>));
	gs_cols=(GaussSeidelRow_M1<double>*)malloc(
		gs_n*sizeof(GaussSeidelRow_M1<double>));
	assert(gs_rows && gs_cols);
	// Allocate nemory for the coefficients and pointers: 
	// Factor 5 because system has max 5 entries per row. 
	size_t need_mem=2*5*(sizeof(double*))*gs_n;
	gs_coeff_buf=(char*)malloc(need_mem);
	assert(gs_coeff_buf);
	char *ptrR=gs_coeff_buf;
	char *ptrC=gs_coeff_buf+need_mem/2;
	for(int i=0; i<gs_n; i++)
	{
		gs_rows[i].x=(double**)ptrR;  ptrR+=5*sizeof(double*);
		gs_cols[i].x=(double**)ptrC;  ptrC+=5*sizeof(double*);
	}
	assert(ptrC==gs_coeff_buf+need_mem);
	
	int n=0;
	for(int y=0; y<height; y++)
	{
		for(int x=0; x<width; x++)
		{
			if(elem(x,y).type==EElement::Fixed)  continue;
			assert(n<gs_n);
			_GSSetupElement(x,y,&gs_rows[n],sp->periodic);
			++n;
		}
	}
	assert(n==gs_n);
	
	n=0;
	for(int x=0; x<width; x++)
	{
		for(int y=0; y<height; y++)
		{
			if(elem(x,y).type==EElement::Fixed)  continue;
			assert(n<gs_n);
			_GSSetupElement(x,y,&gs_cols[n],sp->periodic);
			++n;
		}
	}
	assert(n==gs_n);
	
	gs_sys_dirty=0;
	gs_sys_periodic=sp->periodic;
}


void EStatArea::FreeGSSys()
{
	if(gs_rows)
	{  free(gs_rows);  gs_rows=NULL;  }
	if(gs_cols)
	{  free(gs_cols);  gs_cols=NULL;  }
	if(gs_coeff_buf)
	{  free(gs_coeff_buf);  gs_coeff_buf=NULL;  }
	gs_n=0;
}
