/*
 * solve-tdma.cc
 * 
 * Electrostatic PDE solver Thomas TDMA solver implementation. 
 * 
 * Copyright (c) 2004 by Wolfgang Wieser (wwieser@gmx.de) 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#include <stdio.h>
#include <math.h>
#include <assert.h>

#include "elscore.h"


double EStatArea::_SolveTDM_H(double *diag,double *phi,
	int a,int b,int sy,int syu,int syl,
	double omega,double omega1)
{
	int w1=width-1;
	
	int n=b-a;
	if(syu==-1 || syl==width)  // first or last line
	{
		int syN = syu==-1 ? syl : syu;
		for(int i=0; i<=n; i++)
		{
			diag[i]=3.0;
			phi[i] = elem(i+a,syN).phi;
		}
	}
	else
	{
		for(int i=0; i<=n; i++)
		{
			diag[i]=4.0;
			phi[i] = elem(i+a,syu).phi + elem(i+a,syl).phi;
		}
	}
	if(a==0)  // free left border
	{  diag[0]-=1.0;  }
	else
	{  phi[0] += elem(a-1,sy).phi;  }
	if(b==w1)  // free right border
	{  diag[n]-=1.0;  }
	else
	{  phi[n] += elem(b+1,sy).phi;  }
	
	// Solve: 
	SolveTridiag_M1_TDMA(n,diag,phi,phi);
	
	// Write back: 
	double diffmax=0.0;
	for(int i=0; i<=n; i++)
	{
		double &ep=elem(i+a,sy).phi;
		double diff=fabs(ep-phi[i]);
		ep = phi[i]*omega + omega1*ep;
		if(diff>diffmax)  diffmax=diff;
	}
	return(diffmax);
}

double EStatArea::_SolveTDM_V(double *diag,double *phi,
	int a,int b,int sx,int sxu,int sxl,
	double omega,double omega1)
{
	int h1=height-1;
	
	int n=b-a;
	if(sxu==-1 || sxl==height)  // first or last line
	{
		int sxN = sxu==-1 ? sxl : sxu;
		for(int i=0; i<=n; i++)
		{
			diag[i]=3.0;
			phi[i] = elem(sxN,i+a).phi;
		}
	}
	else
	{
		for(int i=0; i<=n; i++)
		{
			diag[i]=4.0;
			phi[i] = elem(sxu,i+a).phi + elem(sxl,i+a).phi;
		}
	}
	if(a==0)  // free left border
	{  diag[0]-=1.0;  }
	else
	{  phi[0] += elem(sx,a-1).phi;  }
	if(b==h1)  // free right border
	{  diag[n]-=1.0;  }
	else
	{  phi[n] += elem(sx,b+1).phi;  }
	
	// Solve: 
	SolveTridiag_M1_TDMA(n,diag,phi,phi);
	
	// Write back: 
	double diffmax=0.0;
	for(int i=0; i<=n; i++)
	{
		double &ep=elem(sx,i+a).phi;
		double diff=fabs(ep-phi[i]);
		ep = phi[i]*omega + omega1*ep;
		if(diff>diffmax)  diffmax=diff;
	}
	return(diffmax);
}


double EStatArea::TDMA_Iterate(SolverParams *sp,
	int sdir,int sinc)
{
	assert(width>=2 && height>=2);
	
	if(chunks_dirty)
	{  ComputeChunkLines();  }
	assert(ck_rows && ck_cols);
	// Okay, now we have all the chunk lines. 
	
	int maxsize = width<height ? height : width;
	// Diagonal elements: 
	double diag[maxsize];
	// Potential values: 
	double phi[maxsize];
	
	double diffmax=0.0;
	
	// Overrelaxation: 
	double omega,omega1;
	if(sp->omega<0.0)
	{  omega=1.0;  omega1=0.0;  }
	else
	{  omega=sp->omega;  omega1=1.0-omega;  }
	
	// TDMA version cannot handle periodic BC because that would make 
	// the linear systems non-diagonal (more exactly this could produce 
	// linear systems which are tridiagonal with but additionally with 
	// the right top and left bottom element non-zero. 
	assert(!sp->periodic);
	if(sdir==0)
	{
		// Sweep row-by-row: 
		int w1=width-1;
		for(int sy=sinc<0 ? height-1 : 0,
	    	 syend=sinc<0 ? -1 : height; 
			sy!=syend; sy+=sinc)
		{
			ChunkLine *cl=&ck_rows[sy];

			// Upper and lower neightbour lines. 
			int syu = sy-1;
			int syl = sy+1;

			if(cl->nchunks<0)
			{
				if(cl->nchunks==-1)  // all floating
				{
					double diff=_SolveTDM_H(diag,phi,0,w1,sy,syu,syl,
						omega,omega1);
					if(diff>diffmax)  diffmax=diff;
				}
				// else -2 -> all fixed; simply skip. 
				continue;
			}

			// cl->nchunks>=0
			for(int c=0; c<cl->nchunks; c++)
			{
				int a=cl->chunk[c].a;
				int b=cl->chunk[c].b;
				if(a<=b)
				{
					// Normal, non-wrapping chunk. 
					double diff=_SolveTDM_H(diag,phi,a,b,sy,syu,syl,
						omega,omega1);
					if(diff>diffmax)  diffmax=diff;
				}
				else  // a>b
				{
					// A wrapping chunk: 
					double diff=_SolveTDM_H(diag,phi,0,b,sy,syu,syl,
						omega,omega1);
					if(diff>diffmax)  diffmax=diff;
					diff=_SolveTDM_H(diag,phi,a,w1,sy,syu,syl,
						omega,omega1);
					if(diff>diffmax)  diffmax=diff;
				}
			}
		}
	}
	else
	{
		// Sweep column-by-column: 
		int h1=height-1;
		for(int sx=sinc<0 ? width-1 : 0,
	    	 sxend=sinc<0 ? -1 : width; 
			sx!=sxend; sx+=sinc)
		{
			ChunkLine *cl=&ck_cols[sx];

			// Upper and lower neightbour lines. 
			int sxu = sx-1;
			int sxl = sx+1;

			if(cl->nchunks<0)
			{
				if(cl->nchunks==-1)  // all floating
				{
					double diff=_SolveTDM_V(diag,phi,0,h1,sx,sxu,sxl,
						omega,omega1);
					if(diff>diffmax)  diffmax=diff;
				}
				// else -2 -> all fixed; simply skip. 
				continue;
			}

			// cl->nchunks>=0
			for(int c=0; c<cl->nchunks; c++)
			{
				int a=cl->chunk[c].a;
				int b=cl->chunk[c].b;
				if(a<=b)
				{
					// Normal, non-wrapping chunk. 
					double diff=_SolveTDM_V(diag,phi,a,b,sx,sxu,sxl,
						omega,omega1);
					if(diff>diffmax)  diffmax=diff;
				}
				else  // a>b
				{
					// A wrapping chunk: 
					double diff=_SolveTDM_V(diag,phi,0,b,sx,sxu,sxl,
						omega,omega1);
					if(diff>diffmax)  diffmax=diff;
					diff=_SolveTDM_V(diag,phi,a,h1,sx,sxu,sxl,
						omega,omega1);
					if(diff>diffmax)  diffmax=diff;
				}
			}
		}
	}
	
	return(diffmax);
}


void EStatArea::ComputeChunkLines()
{
	fprintf(stderr,"(CCL)");
	
	FreeChunkLines(ck_rows,height);
	FreeChunkLines(ck_cols,width);
	
	if(!ck_rows)
	{
		ck_rows=(ChunkLine*)malloc(height*sizeof(ChunkLine));
		assert(ck_rows);
	}
	if(!ck_cols)
	{
		ck_cols=(ChunkLine*)malloc(width*sizeof(ChunkLine));
		assert(ck_cols);
	}
	
	for(int y=0; y<height; y++)
	{
		ChunkLine *cl=&ck_rows[y];
		cl->nchunks=1;  // Correct. 
		
		// Count the number of needed chunks: 
		EElement::Type t=elem(0,y).type;
		for(int x=1; x<width; x++)
		{
			EElement::Type t2=elem(x,y).type;
			//fprintf(stderr,"<%d> ",t2);
			if(t2!=t)
			{  ++cl->nchunks;  t=t2;  }
		}
		//fprintf(stderr,"  <%d, %d>\n",y,cl->nchunks);
		cl->nchunks/=2;
		
		if(!cl->nchunks)
		{
			cl->chunk=NULL;
			
			// All fixed or all floating: 
			if(t==EElement::Fixed)  cl->nchunks=-2;
			else  cl->nchunks=-1;
		}
		else
		{
			cl->chunk=(LChunk*)malloc(cl->nchunks*sizeof(LChunk));
			assert(cl->chunk);
			
			int ii=0;
			EElement::Type t=elem(0,y).type;
			if(t==EElement::Fixed)
			{
				// First is fixed. Skip them. 
				int x=1;
				for(;;)
				{
					while(x<width && elem(x,y).type==EElement::Fixed)  ++x;
					if(x==width)  break;
					cl->chunk[ii].a=x;
					// Walk over the floating: 
					while(x<width && elem(x,y).type==EElement::Floating)  ++x;
					if(x==width)
					{  cl->chunk[ii++].b=x-1;  break;  }
					cl->chunk[ii++].b=x-1;
				}
			}
			else
			{
				// First is floating. 
				int x=1;
				for(;;)
				{
					// Walk over the floating: 
					while(x<width && elem(x,y).type==EElement::Floating)  ++x;
					assert(x!=width);
					cl->chunk[ii++].b=x-1;
					// Walk over the fixed: 
					while(x<width && elem(x,y).type==EElement::Fixed)  ++x;
					if(x==width)
					{  cl->chunk[0].a=0;  break;  }
					if(ii==cl->nchunks)
					{  cl->chunk[0].a=x;  break;  }
					cl->chunk[ii].a=x;
				}
			}
			assert(ii==cl->nchunks);
		}
	}
	
	// Okay, and now the same procedure in x direction: 
	for(int x=0; x<width; x++)
	{
		ChunkLine *cl=&ck_cols[x];
		cl->nchunks=1;  // Correct. 
		
		// Count the number of needed chunks: 
		EElement::Type t=elem(x,0).type;
		for(int y=1; y<height; y++)
		{
			EElement::Type t2=elem(x,y).type;
			//fprintf(stderr,"<%d> ",t2);
			if(t2!=t)
			{  ++cl->nchunks;  t=t2;  }
		}
		//fprintf(stderr,"  <%d, %d>\n",y,cl->nchunks);
		cl->nchunks/=2;
		
		if(!cl->nchunks)
		{
			cl->chunk=NULL;
			
			// All fixed or all floating: 
			if(t==EElement::Fixed)  cl->nchunks=-2;
			else  cl->nchunks=-1;
		}
		else
		{
			cl->chunk=(LChunk*)malloc(cl->nchunks*sizeof(LChunk));
			assert(cl->chunk);
			
			int ii=0;
			EElement::Type t=elem(x,0).type;
			if(t==EElement::Fixed)
			{
				// First is fixed. Skip them. 
				int y=1;
				for(;;)
				{
					while(y<height && elem(x,y).type==EElement::Fixed)  ++y;
					if(y==height)  break;
					cl->chunk[ii].a=y;
					// Walk over the floating: 
					while(y<height && elem(x,y).type==EElement::Floating)  ++y;
					if(y==height)
					{  cl->chunk[ii++].b=y-1;  break;  }
					cl->chunk[ii++].b=y-1;
				}
			}
			else
			{
				// First is floating. 
				int y=1;
				for(;;)
				{
					// Walk over the floating: 
					while(y<height && elem(x,y).type==EElement::Floating)  ++y;
					assert(y!=height);
					cl->chunk[ii++].b=y-1;
					// Walk over the fixed: 
					while(y<height && elem(x,y).type==EElement::Fixed)  ++y;
					if(y==height)
					{  cl->chunk[0].a=0;  break;  }
					if(ii==cl->nchunks)
					{  cl->chunk[0].a=y;  break;  }
					cl->chunk[ii].a=y;
				}
			}
			assert(ii==cl->nchunks);
		}
	}
	
	#if 0
	for(int y=0; y<height; y++)
	{
		ChunkLine *cl=&ck_rows[y];
		fprintf(stderr,"row[%d]: %d chks:",y,cl->nchunks);
		for(int c=0; c<cl->nchunks; c++)
			fprintf(stderr," %d..%d",cl->chunk[c].a,cl->chunk[c].b);
		fprintf(stderr,"\n");
	}
	for(int x=0; x<width; x++)
	{
		ChunkLine *cl=&ck_cols[x];
		fprintf(stderr,"col[%d]: %d chks:",x,cl->nchunks);
		for(int c=0; c<cl->nchunks; c++)
			fprintf(stderr," %d..%d",cl->chunk[c].a,cl->chunk[c].b);
		fprintf(stderr,"\n");
	}
	#endif
	
	chunks_dirty=0;
}


void EStatArea::FreeChunkLines(ChunkLine *cl_array,int size)
{
	if(!cl_array)  return;
	
	// Free the chunk lines: 
	for(int i=0; i<size; i++)
	{
		ChunkLine *cl=&cl_array[i];
		if(cl->chunk)
		{  free(cl->chunk);  cl->chunk=NULL;  }
	}
}
