/*
 * areawindow.cc
 * 
 * ELStat's area window. 
 * 
 * Copyright (c) 2002--2004 by Wolfgang Wieser (wwieser@gmx.de) 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#include "areawindow.h"

#include <Qt/qpushbutton.h>
#include <Qt/qcheckbox.h>
#include <Qt/qlayout.h>
#include <Qt/qlabel.h>
#include <Qt/qvalidator.h>

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>


void AreaWindow::_Restart()
{
	if(running==1)
	{
		startTimer(0);
		running=2;
	}
}

void AreaWindow::restart_solver()
{
	_Restart();
}


void AreaWindow::resizeEvent(QResizeEvent *)
{
	int nw=size().width();
	int nh=size().height();
	
	int aww=Cell2WinX(area.Width())-Cell2WinX(0);
	int awh=Cell2WinY(area.Height())-Cell2WinY(0);
	bool can_scroll_x=(aww>nw);
	bool can_scroll_y=(awh>nh);
	
	if(!can_scroll_x)
	{  cellx0=-(nw-aww)/2;  }
	else
	{
		if(Cell2WinX(0)>0)
		{  cellx0=0;  }
		else if(Cell2WinX(area.Width())<nw)
		{  cellx0=area.Width()*cellw-nw;  }
	}
	
	if(!can_scroll_y)
	{  celly0=-(nh-awh)/2;  }
	else
	{
		if(Cell2WinY(0)>0)
		{  celly0=0;  }
		else if(Cell2WinY(area.Height())<nh)
		{  celly0=area.Height()*cellh-nh;  }
	}
	
	
	//fprintf(stderr,"xy0=%d,%d\n",cellx0,celly0);
}


void AreaWindow::paintEvent(QPaintEvent *ev)
{
	if(!isVisible())  return;
	
	#if 0
	QRect qr=ev->rect();
	int px=qr.x();
	int py=qr.y();
	int pw=qr.width();
	int ph=qr.height();
	#endif
	
	Update(1);
}


void AreaWindow::mouseMoveEvent(QMouseEvent *ev)
{
	if(!(ev->state()&(LeftButton|RightButton)))
	{  return;  }
	
	if(area_dragging)
	{
		_ProcessDrag(ev->x(),ev->y(),0);
	}
	if(drawing)
	{
		_ProcessDrawing(ev->x(),ev->y(),0);
	}
}

void AreaWindow::mousePressEvent(QMouseEvent *ev)
{
	if(ev->button()&RightButton)
	{
		areadrag_x0=ev->x();
		areadrag_y0=ev->y();
		areadrag_cx0=cellx0;
		areadrag_cy0=celly0;
		area_dragging=1;
	}
	if(ev->button()&LeftButton)
	{
		draw_x0=Win2CellX(ev->x());
		draw_y0=Win2CellY(ev->y());
		draw_prev_x=draw_x0;
		draw_prev_y=draw_y0;
		emit get_draw_params(&drawpar);
		assert(drawpar.valid);
		drawing=1;
		_ProcessDrawing(ev->x(),ev->y(),0);
	}
}

void AreaWindow::mouseReleaseEvent(QMouseEvent *ev)
{
	if(ev->button()&RightButton && area_dragging)
	{
		_ProcessDrag(ev->x(),ev->y(),1);
		area_dragging=0;
	}
	if(ev->button()&LeftButton && drawing)
	{
		_ProcessDrawing(ev->x(),ev->y(),1);
		drawing=0;
	}
}


void AreaWindow::keyPressEvent(QKeyEvent *ev)
{
	switch(ev->key())
	{
		case Key_Plus:
			if(cellw<32 && cellh<32)
			{
				++cellw;
				++cellh;
				resizeEvent(NULL);
				Update(1);
			}
			break;
		case Key_Minus:
			if(cellw>1 && cellh>1)
			{
				--cellw;
				--cellh;
				resizeEvent(NULL);
				Update(1);
			}
			break;
		default:
			emit key_press_event(ev);
			break;
	}
}

void AreaWindow::keyReleaseEvent(QKeyEvent *ev)
{
	switch(ev->key())
	{
		case Key_Plus:
		case Key_Minus:
			break;
		default:
			emit key_release_event(ev);
			break;
	}
}


void AreaWindow::timerEvent(QTimerEvent *)
{
	assert(running);
	
	// Query parameters for the run: 
	curr_sp.valid=0;
	emit get_solver_params(&curr_sp);
	assert(curr_sp.valid);
	
	int rv=area.RunSolver(&curr_sp);
	switch(rv)
	{
		case 0:
			running=2;
			break;
		case 1:
		case -1:
			killTimers();
			running=1;
			emit pde_converged(rv);
			break;
	}
	
	Update(0);
	fprintf(stderr,running==2 ? "." : "|");
}


void AreaWindow::_ProcessDrag(int mx,int my,bool end)
{
	int dx=mx-areadrag_x0;
	int dy=my-areadrag_y0;
	
	int nw=size().width();
	int nh=size().height();
	
	int aww=Cell2WinX(area.Width())-Cell2WinX(0);
	int awh=Cell2WinY(area.Height())-Cell2WinY(0);
	bool can_scroll_x=(aww>nw);
	bool can_scroll_y=(awh>nh);
	if(!can_scroll_x)  dx=0;
	if(!can_scroll_y)  dy=0;
	if(!dx && !dy)  return;
	
	cellx0=areadrag_cx0-dx;
	celly0=areadrag_cy0-dy;
	
	Update(1);
}


void AreaWindow::_RedrawShape(int cx,int cy)
{
	int x0 = draw_x0<cx ? draw_x0 : cx;
	int y0 = draw_y0<cy ? draw_y0 : cy;
	int w=abs(draw_x0-cx);
	int h=abs(draw_y0-cy);
	int px=Cell2WinX(x0)+cellw/2;
	int py=Cell2WinY(y0)+cellh/2;
	int pw=w*cellw;
	int ph=h*cellh;
	//fprintf(stderr,"shape(%d %d %d %d)\n",x0,y0,w,h);
	if(drawpar.shape==DrawParams::Rectangle)
	{  xor_pnt.DrawRectangle(px,py,pw,ph);  }
	else if(drawpar.shape==DrawParams::Ellipse)
	{  xor_pnt.DrawArc(px,py,pw,ph);  }
}


void AreaWindow::_ProcessDrawing(int mx,int my,bool end)
{
	int cx=Win2CellX(mx);
	int cy=Win2CellY(my);
	
	switch(drawpar.shape)
	{
		case DrawParams::Point:
		{
			_DoDrawPoint(cx,cy);
		}	break;
		case DrawParams::Rectangle:
		case DrawParams::Ellipse:
		{
			if((cx!=draw_prev_x || cy!=draw_prev_y) || end)
			{
				_RedrawShape(draw_prev_x,draw_prev_y);
				draw_prev_x=cx;
				draw_prev_y=cy;
				if(!end)
				{  _RedrawShape(draw_prev_x,draw_prev_y);  }
			}
			if(end)
			{
				int x0 = draw_x0<cx ? draw_x0 : cx;
				int y0 = draw_y0<cy ? draw_y0 : cy;
				int w=abs(draw_x0-cx);
				int h=abs(draw_y0-cy);
				if(drawpar.shape==DrawParams::Rectangle)
				{
					for(int x=0; x<=w; x++)
					{
						_DoDrawPoint(x+x0,y0);
						_DoDrawPoint(x+x0,y0+h);
					}
					for(int y=1; y<h; y++)
					{
						_DoDrawPoint(x0,y+y0);
						_DoDrawPoint(x0+w,y+y0);
					}
				}
				else if(drawpar.shape==DrawParams::Ellipse)
				{
					// This ellipse drawing algorithm is far from perfect. 
					int numi=int(2.0*M_PI/atan(1.0/(w>h ? w : h))+1.0);
					//numi*=2;
					for(int i=0; i<numi; i++)
					{
						double phi=2.0*M_PI*i/numi;
						int xx=int(0.5*w*(1.0+cos(phi))+0.5);
						int yy=int(0.5*h*(1.0+sin(phi))+0.5);
						_DoDrawPoint(xx+x0,yy+y0);
					}
				}
			}
		}	break;
		default: assert(0);
	}
}


void AreaWindow::_DoDrawPoint(int cx,int cy)
{
	if(cx<0 || cy<0 || cx>=area.Width() || cy>=area.Height())  return;
	
	int mmchg=area.UpdateElement(cx,cy,drawpar.phi,drawpar.type);
	if(mmchg)
	{
		UpdateElem(cx,cy);  // Must be done even if we Update(0) afterwards
		if(mmchg==2)        // because Update(0) does no erase(). 
		{  Update(0);  }
	}
	
	//fprintf(stderr,"draw(%d,%d,%d,%g)=%d\n",
	//	cx,cy,drawpar.type,drawpar.phi,mmchg);
	
	_Restart();
}


void AreaWindow::StartStopSolver()
{
	if(running)
	{
		killTimers();
		running=0;
	}
	else
	{
		running=2;
		startTimer(0);
	}
	
	emit runstate_change(running);
}


void AreaWindow::fillin_area_setup(AreaSetupPars *spar)
{
	area.GetAreaSetup(spar);
}


void AreaWindow::area_setup_changed(AreaSetupPars *spar)
{
	area.SetupArea(spar);
	resizeEvent(NULL);
	Update(1);
	_Restart();
}

void AreaWindow::ResetArea(EElement::Type back)
{
	area.Reset(back);
	resizeEvent(NULL);
	Update(1);
	_Restart();
}

void AreaWindow::SetAllFloating(double phi)
{
	area.SetAllFloating(phi);
	Update(1);
	_Restart();
}


void AreaWindow::_InitColors()
{
	double minp=area.MinPhi();
	double maxp=area.MaxPhi();
	double delta=maxp-minp;
	if(delta<1e-3)
	{  maxp+=1e-3;  delta=maxp-minp;  }
	
	//fprintf(stderr,"COLORS: min=%g, max=%g, delta=%g\n",
	//	minp,maxp,delta);
	
	phi_off=minp;
	phi_scale=(ncolors-1)/delta;
	
	// Compute the color values: 
	for(int i=0; i<ncolors; i++)
	{
		double p=double(i)/double(ncolors-1);
		float r = sin(p*M_PI/2.0);
		float g = 0.0;
		float b = cos(p*M_PI/2.0);
		u_int32_t color=colsrc.color(r,g,b);
		rpnt[i]->SetForeground(color);
		rpnt[i]->SetBackground(color);
		ppnt[i]->SetForeground(color);
		ppnt[i]->SetBackground(color);
	}
	
	emit colors_changed();
}


// Re-draw complete screen: 
void AreaWindow::Update(bool do_erase)
{
	//fprintf(stderr,"COMPLETE UPDATE(%d,%d,%d,%d)\n",cellx0,celly0,cellw,cellh);
	
	_InitColors();
	
	int winw=size().width();
	int winh=size().height();
	
	int areaw=area.Width();
	int areah=area.Height();
	
	// Complete update: 
	int ax0=Win2CellX(0);
	int ay0=Win2CellY(0);
	int ax1=Win2CellX(winw+cellw);
	int ay1=Win2CellY(winh+cellh);
	
	//fprintf(stderr,"x=%d,%d... ",ax0,ax1);
	
	if(ax0<0)  ax0=0;
	if(ay0<0)  ay0=0;
	if(ax1<ax0)  ax1=ax0;
	if(ay1<ay0)  ay1=ay0;
	
	if(ax1>=areaw)  ax1=areaw-1;
	if(ay1>=areah)  ay1=areah-1;
	if(ax0>ax1)  ax0=ax1;
	if(ay0>ay1)  ay0=ay1;
	
	//fprintf(stderr,"x2=%d,%d\n",ax0,ax1);
	
	if(do_erase)
	{  erase();  }
	
	// Black surrounding box: 
	if(_MustDrawGrid())
	{  black_pnt.DrawRectangle(
		Cell2WinX(0),Cell2WinY(0),
		Cell2WinX(areaw)-Cell2WinX(0),Cell2WinY(areah)-Cell2WinY(0));  }
	else
	{  black_pnt.DrawRectangle(
		Cell2WinX(0)-1,Cell2WinY(0)-1,
		Cell2WinX(areaw)-Cell2WinX(0)+1,Cell2WinY(areah)-Cell2WinY(0)+1);  }
	
	// Draw grid: 
	if(_MustDrawGrid())
	{
		int sl=Cell2WinX(0)+1;      if(sl<0)  sl=0;  if(sl>=winw)  sl=winw;
		int el=Cell2WinX(areaw)-1;  if(el<0)  el=0;  if(el>=winw)  el=winw;
		for(int ay=ay0+1; ay<=ay1; ay++)
		{
			int wc=Cell2WinY(ay);
			grid_pnt.DrawLine(sl,wc,el,wc);
		}
		sl=Cell2WinY(0)+1;      if(sl<0)  sl=0;  if(sl>=winh)  sl=winh;
		el=Cell2WinY(areah)-1;  if(el<0)  el=0;  if(el>=winh)  el=winh;
		for(int ax=ax0+1; ax<=ax1; ax++)
		{
			int wc=Cell2WinX(ax);
			grid_pnt.DrawLine(wc,sl,wc,el);
		}
	}
	
	bool use_ppnt=(cellw==1 && cellh==1);
	for(int ay=ay0; ay<=ay1; ay++)
	{
		assert(ay>=0 && ay<areah);
		
		for(int ax=ax0; ax<=ax1; ax++)
		{
			assert(ax>=0 && ax<areaw);
			
			const EElement *e=&area.Elem(ax,ay);
			if(e->type==EElement::Floating)
			{
				if(use_ppnt)
				{
					ppnt.AddPoint(ColIdx(e->phi),
						Cell2WinX(ax),Cell2WinY(ay));
				}
				else if(_MustDrawGrid())
				{
					rpnt.AddRect(ColIdx(e->phi),
						Cell2WinX(ax)+1,Cell2WinY(ay)+1,
						cellw-1,cellh-1);
				}
				else
				{
					rpnt.AddRect(ColIdx(e->phi),
						Cell2WinX(ax),Cell2WinY(ay),
						cellw,cellh);
				}
			}
			else
			{
				// Fixed. 
				if(_MustDrawGrid())
				{
					rpnt.AddRect(ColIdx(e->phi),
						Cell2WinX(ax)+2,Cell2WinY(ay)+2,
						cellw-3,cellh-3);
				}
				else if(cellw>=3 && cellh>=3)
				{
					rpnt.AddRect(ColIdx(e->phi),
						Cell2WinX(ax)+1,Cell2WinY(ay)+1,
						cellw-2,cellh-2);
				}
			}
		}
	}
	
	if(use_ppnt)  ppnt.Flush();
	else  rpnt.Flush();
	
	XSync(rpnt[0]->display(),False);
	
	if(drawing)
	{  _RedrawShape(draw_prev_x,draw_prev_y);  }
}


void AreaWindow::UpdateElem(int ax,int ay)
{
	const EElement *e=&area.Elem(ax,ay);
	int cidx=ColIdx(e->phi);
	bool use_ppnt=(cellw==1 && cellh==1);
	if(e->type==EElement::Floating)
	{
		if(use_ppnt)
		{
			ppnt.AddPoint(cidx,
				Cell2WinX(ax),Cell2WinY(ay));
		}
		else if(_MustDrawGrid())
		{
			rpnt.AddRect(cidx,
				Cell2WinX(ax)+1,Cell2WinY(ay)+1,
				cellw-1,cellh-1);
		}
		else
		{
			rpnt.AddRect(cidx,
				Cell2WinX(ax),Cell2WinY(ay),
				cellw,cellh);
		}
	}
	else
	{
		// Fixed. 
		if(_MustDrawGrid())
		{
			white_pnt.FillRectangle(
				Cell2WinX(ax)+1,Cell2WinY(ay)+1,
				cellw-1,cellh-1);
			rpnt.AddRect(cidx,
				Cell2WinX(ax)+2,Cell2WinY(ay)+2,
				cellw-3,cellh-3);
		}
		else 
		{
			white_pnt.FillRectangle(
				Cell2WinX(ax),Cell2WinY(ay),
				cellw,cellh);
			if(cellw>=3 && cellh>=3)
			{
				rpnt.AddRect(cidx,
					Cell2WinX(ax)+1,Cell2WinY(ay)+1,
					cellw-2,cellh-2);
			}
		}
	}
	
	if(use_ppnt)  ppnt.Flush(cidx);
	else  rpnt.Flush(cidx);
	
	_Restart();
}


void AreaWindow::change_ncolors(int ncol)
{
	if(ncol==ncolors || ncol<2)  return;
	
	ncolors=ncol;
	
	rpnt.Setup(ncolors);
	ppnt.Setup(ncolors);
	
	QPainter pnt(this);
	pnt.setBrush(QBrush(Qt::white,Qt::SolidPattern));
	for(int i=0; i<ncolors; i++)
	{
		rpnt.Setup(i,&pnt,32);
		ppnt.Setup(i,&pnt,32);
	}
	
	Update(1);
}


void AreaWindow::Map()
{
	if(isVisible())
	{
		hide();
		#warning disable sim updates?
	}
	else
	{
		show();
	}
}


AreaWindow::AreaWindow(QWidget *_parent,const char *name,WFlags f) : 
	QWidget(_parent,name,f),
	area(),
	ncolors(0),
	colsrc(this),
	rpnt(1),
	ppnt(1),
	grid_pnt(this),
	white_pnt(this),
	black_pnt(this),
	xor_pnt(this)
{
	cellx0=0;
	celly0=0;
	cellw=8;
	cellh=8;
	
	phi_off=0.0;
	phi_scale=1.0/100.0;
	
	area_dragging=0;
	drawing=0;
	
	running=0;
	AreaSetupPars asp;
	asp.width=50;
	asp.height=50;
	asp.background=EElement::Floating;
	asp.valid=1;
	area.SetupArea(&asp);
	
	resize(408,408);
	
	QPainter pnt(this);
	pnt.setBrush(QBrush(Qt::white,Qt::SolidPattern));
	white_pnt.QCopyGC(pnt,1);
	xor_pnt.QCopyGC(pnt,1);
	xor_pnt.SetFunction(GXxor);
	
	change_ncolors(256);
	
	pnt.setBrush(QBrush(QColor(0xdd,0xdd,0xdd),Qt::SolidPattern));
	grid_pnt.QCopyGC(pnt,1);
	
	pnt.setBrush(QBrush(Qt::black,Qt::SolidPattern));
	black_pnt.QCopyGC(pnt,1);
	
	setBackgroundColor(Qt::white);
	
	setCaption("Electrostatic PDE Simulator");
}


AreaWindow::~AreaWindow()
{
}

/******************************************************************************/

void ColorGradientDisplay::_Redraw()
{
	int w=width()-2;
	int h=height();
	
	QPainter qpnt(this);
	qpnt.setBrush(QBrush(Qt::white,Qt::SolidPattern));
	pnt.QCopyGC(qpnt,1);
	int nc=areawin->ppnt.GetArraySize();
	int oldj=-1;
	
	for(int x=0; x<w; x++)
	{
		double p=double(x)/(w);
		int j=int(p*(nc));
		if(j>=nc)  j=nc-1;
		if(j!=oldj)
		{
			XGCValues gcv;
			XGetGCValues(areawin->ppnt[j]->display(),
				areawin->ppnt[j]->get_gc(),
				GCForeground,&gcv);
			pnt.SetForeground(gcv.foreground);
		}
		pnt.DrawLine(x+1,0,x+1,h);
	}
	
	qpnt.setBrush(QBrush(Qt::black,Qt::SolidPattern));
	pnt.QCopyGC(qpnt,1);
	pnt.DrawRectangle(0,0,w,h);
	
	char tmp[32];
	
	//qpnt.setBrush(QBrush(Qt::white,Qt::SolidPattern));
	//pnt.QCopyGC(qpnt,1);
	//pnt.SetFunction(GXxor);
	
	qpnt.setBrush(QBrush(Qt::white,Qt::SolidPattern));
	pnt.QCopyGC(qpnt,1);
	snprintf(tmp,32,"%g",areawin->phi_off*1000.0);
	pnt.DrawString(1,h-2,tmp);
	
	double mval=(areawin->phi_off+(nc-1)/areawin->phi_scale);
	if(fabs(mval)<1e-4)  mval=0.0;
	snprintf(tmp,32,"%g",mval*1000.0);
	//XGCValues gcv;
	//XGetGCValues(pnt.display(),pnt.get_gc(),GCFont,&gcv);
	//XFontStruct *fstr=XQueryFont(pnt.display(),gcv.font);
	//assert(fstr);
	//pnt.DrawString(w-XTextWidth(fstr,tmp,strlen(tmp))-1,h-2,tmp);
	//XFreeFontInfo(NULL,fstr,0);
	pnt.DrawString(w-strlen(tmp)*6-1,h-2,tmp);
	
	pnt.SetFunction(GXcopy);
}


void ColorGradientDisplay::update()
{
	_Redraw();
}


void ColorGradientDisplay::paintEvent(QPaintEvent *)
{
	_Redraw();
}


void ColorGradientDisplay::resizeEvent(QResizeEvent *)
{
	
}


ColorGradientDisplay::ColorGradientDisplay(AreaWindow *_areawin,
	QWidget *_parent,const char *name,WFlags f) : 
	QWidget(_parent,name,f),
	pnt(this)
{
	areawin=_areawin;
	setMinimumHeight(20);
	
	connect(areawin,SIGNAL(colors_changed()),this,SLOT(update()));
}

ColorGradientDisplay::~ColorGradientDisplay()
{
}
