/*
 * areawindow.h
 * 
 * ELStat's area window. 
 * 
 * Copyright (c) 2002--2004 by Wolfgang Wieser (wwieser@gmx.de) 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#ifndef _ELSTAT_QTXGUI_AREAWINDOW_H_
#define _ELSTAT_QTXGUI_AREAWINDOW_H_ 1

#include "ecore/elscore.h"
#include "areasetup.h"

#include <Qt/qwidget.h>

#include <QTX/bufrectpainterarray.h>
#include <QTX/bufpointpainterarray.h>
#include <QTX/rgbcolorsource.h>


class MainWindow;
struct AreaSetupPars;


struct DrawParams
{
	enum Shape
	{
		Point,
		Rectangle,
		Ellipse,
	} shape;
	double phi;
	EElement::Type type;
	bool valid;
	
	DrawParams() { valid=0; }
};


class ColorGradientDisplay;

class AreaWindow : public QWidget
{
	Q_OBJECT
	friend class ColorGradientDisplay;
	private:
		EStatArea area;
		
		int running;
		SolverParams curr_sp;
		
		// Offset and zoom state: 
		int cellx0,celly0;  // in window pixels
		int cellw,cellh;   // in window pixels
		
		// Number of colors for potential values: 
		int ncolors;
		
		// To compute color values: 
		QTX::RGBColorSource colsrc;
		double phi_off;
		double phi_scale;
		
		QTX::BufferedRectPainterArray rpnt;
		QTX::BufferedPointPainterArray ppnt;
		
		QTX::XPainter grid_pnt;
		QTX::XPainter white_pnt;
		QTX::XPainter black_pnt;
		QTX::XPainter xor_pnt;
		
		// Area dragging: 
		bool area_dragging;
		int areadrag_x0,areadrag_y0;
		int areadrag_cx0,areadrag_cy0;
		// Draw parameters: 
		bool drawing;
		int draw_x0,draw_y0;  // area coos
		int draw_prev_x,draw_prev_y;
		DrawParams drawpar;
		
		// Re-start computation if running is set. 
		void _Restart();
		
		// Re-compute colors depending on the potential scale: 
		void _InitColors();
		
		// Convert element/cell coordinates to window coordinates. 
		inline int Cell2WinX(int cx)
			{  return(cx*cellw-cellx0);  }
		inline int Cell2WinY(int cy)
			{  return(cy*cellh-celly0);  }
		
		// Convert window coordinates to cell coordinates: 
		inline int Win2CellX(int wx)
			{  return((wx+cellx0)/cellw);  }
		inline int Win2CellY(int wy)
			{  return((wy+celly0)/cellh);  }
		
		// Computer color index: 
		inline int ColIdx(double phi)
		{
			int idx=int( (phi-phi_off)*phi_scale + 0.5 );
			if(idx<0)  return(0);
			if(idx>=ncolors)  return(ncolors-1);
			return(idx);
		}
		
		// Do we draw a grid?
		bool _MustDrawGrid() const
			{  return(cellw>=8 && cellh>=8);  }
		
		// Actually draw a point: Does all checks. 
		void _DoDrawPoint(int cx,int cy);
		
		void _RedrawShape(int cx,int cy);
		
		// Mouse actions: 
		void _ProcessDrag(int mx,int my,bool end);
		void _ProcessDrawing(int mx,int my,bool end);
		
		// Events (overriding virtuals): 
		void timerEvent(QTimerEvent *);
		void paintEvent(QPaintEvent *);
		void resizeEvent(QResizeEvent *);
		void mouseMoveEvent(QMouseEvent *);
		void mousePressEvent(QMouseEvent *);
		void mouseReleaseEvent(QMouseEvent *);
		void keyPressEvent(QKeyEvent *);
		void keyReleaseEvent(QKeyEvent *);
		
	signals:
		void pde_converged(int);
		void key_press_event(QKeyEvent *);
		void key_release_event(QKeyEvent *);
		void get_draw_params(DrawParams *);
		void get_solver_params(SolverParams *);
		void runstate_change(int);
		void colors_changed();
	public slots:
		// called to query area setup params: 
		void fillin_area_setup(AreaSetupPars *spar);
		// called when area setup changes: 
		void area_setup_changed(AreaSetupPars *spar);
		// Re-Start solver if running is set. 
		void restart_solver();
		// Change number of colors: 
		void change_ncolors(int ncol);
	public:
		AreaWindow(QWidget *_parent=NULL,const char *name=NULL,WFlags f=0);
		~AreaWindow();
		
		int Running()
			{  return(running);  }
		
		// Show/hide window: 
		void Map();
		
		// Sim control: 
		void StartStopSolver();
		
		void ResetArea(EElement::Type back);
		void SetAllFloating(double phi);
		
		// Re-draw the screen: 
		void Update(bool do_erase);
		// ...and one element: 
		void UpdateElem(int cx,int cy);
};


class ColorGradientDisplay : public QWidget
{
	Q_OBJECT
	private:
		QTX::XPainter pnt;
		AreaWindow *areawin;
		
		void _Redraw();
		
		void paintEvent(QPaintEvent *);
		void resizeEvent(QResizeEvent *);
	public slots:
		void update();
	public:
		ColorGradientDisplay(AreaWindow *areawin,
			QWidget *_parent=NULL,const char *name=NULL,WFlags f=0);
		~ColorGradientDisplay();
};

#endif  /* _ELSTAT_QTXGUI_AREAWINDOW_H_ */
