/*
 * mainwindow.cc
 * 
 * ELStat's main window. 
 * 
 * Copyright (c) 2004--2006 by Wolfgang Wieser (wwieser@gmx.de) 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#include "config.h"

#include "mainwindow.h"

#include <Qt/qpushbutton.h>
#include <Qt/qcheckbox.h>
#include <Qt/qlayout.h>
#include <Qt/qlabel.h>
#include <Qt/qlineedit.h>
#include <Qt/qvalidator.h>
#include <Qt/qspinbox.h>
#include <Qt/qgroupbox.h>
#include <Qt/qmessagebox.h>
#include <Qt/qtabwidget.h>
#include <Qt/qtooltip.h>
#include <Qt/qcombobox.h>

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>


void MainWindow::quit_clicked()
{
	close();
}

void MainWindow::info_clicked()
{
	QMessageBox::about(this,
		"Electro Static Simulator",
		"<center><b>ELStat</b>, version " VERSION "</center>"
		"Electrostatic 2D Laplace PDE solver width GUI.<br><br>"
		"Copyright (c) 2004--2006 by Wolfgang Wieser.<br><br>"
		"This program may be copied unter the terms of the GNU GPL.<br>");
}


void MainWindow::reset_clicked()
{
	areawin.ResetArea(asetup.GetCurrBack());
	
	if(!areawin.isVisible())
	{  areawin.Map();  }
	
	_UpdateValues();
	start_button->setText("&Start");
	start_button_text=0;
}


void MainWindow::zero_clicked()
{
	areawin.SetAllFloating(0.0);
	
	if(!areawin.isVisible())
	{  areawin.Map();  }
	
	_UpdateValues();
}


void MainWindow::start_clicked()
{
	areawin.StartStopSolver();
}


void MainWindow::solver_change(int mode)
{
	if(mode==SolverParams::ThomasTDMA)
	{
		periodic_bc->setChecked(0);
		periodic_bc->setEnabled(0);
		//relax_omega->setEnabled(0);
	}
	else
	{
		periodic_bc->setEnabled(1);
		//relax_omega->setEnabled(1);
	}
}

void MainWindow::pde_converged(int)
{
	
}


void MainWindow::fillin_draw_params(DrawParams *dp)
{
	dp->shape=(DrawParams::Shape)shape_select->currentItem();
	dp->type=(EElement::Type)type_select->currentItem();
	dp->phi=potential_val->value()/1000.0;
	dp->valid=1;
}


void MainWindow::fillin_solver_params(SolverParams *sp)
{
	sp->maxiter=limit_iterations->value();
	sp->eps=eps_mant->value()*pow(10.0,eps_exp->value());
	int ov=relax_omega->value();
	sp->omega=(ov==100 ? -1.0 : ov/100.0);
	sp->method=(SolverParams::Method)solver_mode->currentItem();
	sp->shuffle=(SolverParams::ShuffleMode)shuffle_mode->currentItem();
	sp->timeslice=limit_timeslice->value();  // msec
	sp->periodic=periodic_bc->isChecked();
	
	sp->iter=0;   // Do not change!
	sp->valid=1;  // Do not change!
}


void MainWindow::runstate_change(int running)
{
	fprintf(stderr,"r=%d<\n",running);
	
	// Called by sim window on sim stop/cont. 
	if((running && !start_button_text) ||
	   (!running && start_button_text))
	{
		if(running)
		{
			start_button->setText("&Stop");
			//start_button->setAccel(ALT + Key_S);
			start_button_text=1;
		}
		else
		{
			start_button->setText("&Start");
			//start_button->setAccel(ALT + Key_S);
			start_button_text=0;
		}
	}
}


void MainWindow::_UpdateValues()
{
}

// stat: +1 -> press; -1 -> release
void MainWindow::_HandleKeyEvent(QKeyEvent *ev,int stat)
{
	if(ev->isAutoRepeat() || stat==+1)
	{  ev->ignore();  return;  }
	
	switch(ev->key())
	{
		case Key_Return:
		case Key_Enter:
			break;
		case Key_Q:  quit_clicked();  break;
		case Key_A:  area_button->animateClick();  break;
		case Key_R:  reset_button->animateClick();  break;
		case Key_S:  start_button->animateClick();  break;
		case Key_Z:  zero_button->animateClick();  break;
		default:  ev->ignore();  return;
	}
}

void MainWindow::keyPressEvent(QKeyEvent *ev)
{
	_HandleKeyEvent(ev,+1);
}

void MainWindow::keyReleaseEvent(QKeyEvent *ev)
{
	_HandleKeyEvent(ev,-1);
}

// Called for key presses/releases from area window: 
void MainWindow::key_press_slot(QKeyEvent *ev)
{
	MainWindow::keyPressEvent(ev);
}

void MainWindow::key_release_slot(QKeyEvent *ev)
{
	MainWindow::keyReleaseEvent(ev);
}


void MainWindow::timerEvent(QTimerEvent *)
{
	/* only for setup */
	killTimers();
	
	solver_change(solver_mode->currentItem());
	
	_UpdateValues();
	#if 0
	delay_mode->setCurrentItem(0);
	delay_time->setValue(10);
	delay_change();
	bpropd->windparam_change();
	#endif
	
	resize(minimumSize());
	show();
	
	areawin.Map();
	//reset_clicked();
	
	start_button->setFocus();
}


MainWindow::MainWindow(QWidget *_parent,const char *name,WFlags _f) : 
	QWidget(_parent,name,_f),
	areawin(NULL),   // no parent
	asetup(NULL,NULL,WStyle_Customize | WStyle_DialogBorder)   // no parent
{
	QWidget *wtmp;
	
	connect(&asetup,SIGNAL(change_area_setup(AreaSetupPars*)),
		&areawin,SLOT(area_setup_changed(AreaSetupPars*)));
	connect(&asetup,SIGNAL(get_area_setup(AreaSetupPars*)),
		&areawin,SLOT(fillin_area_setup(AreaSetupPars*)));
	
	connect(&areawin,SIGNAL(pde_converged(int)),
		this,SLOT(pde_converged(int)));
	connect(&areawin,SIGNAL(key_press_event(QKeyEvent *)),
		this,SLOT(key_press_slot(QKeyEvent *)));
	connect(&areawin,SIGNAL(key_release_event(QKeyEvent *)),
		this,SLOT(key_release_slot(QKeyEvent *)));
	connect(&areawin,SIGNAL(get_draw_params(DrawParams *)),
		this,SLOT(fillin_draw_params(DrawParams *)));
	connect(&areawin,SIGNAL(get_solver_params(SolverParams *)),
		this,SLOT(fillin_solver_params(SolverParams *)));
	connect(&areawin,SIGNAL(runstate_change(int)),
		this,SLOT(runstate_change(int)));
	
	/******** HEAD ********/
	QWidget *parent=this;
	QPushButton *quit_button=new QPushButton("&Quit",parent);
	connect(quit_button,SIGNAL(clicked()),this,SLOT(quit_clicked()));
	
	QPushButton *info_button=new QPushButton("Info",parent);
	connect(info_button,SIGNAL(clicked()),this,SLOT(info_clicked()));
	
	QHBoxLayout *head_layout=new QHBoxLayout();
	  head_layout->addWidget(quit_button);
	  head_layout->addSpacing(2);
	  head_layout->addWidget(info_button);
	
	
	int margin=5;
	int spacing=5;
	/******** PDE SOLVER ********/
	QGroupBox *solver_group=new QGroupBox("PDE Solver",this);
	parent=solver_group;
	
	limit_iterations=new QSpinBox(0,999999,10,parent);
	  limit_iterations->setValue(0);
	  limit_iterations->setSpecialValueText("[no limit]");
	
	limit_timeslice=new QSpinBox(0,1000,10,parent);
	  limit_timeslice->setValue(100);
	
	relax_omega=new QSpinBox(100,199,10,parent);
	  relax_omega->setValue(170);
	  relax_omega->setSpecialValueText("[off]");
	  relax_omega->setSuffix("%");
	
	QString _solver_explanation(
		"PDE solve algorithm to use.");
	solver_mode=new QComboBox(parent);
	  solver_mode->insertItem("Gauss-Seidel",SolverParams::GaussSeidel);
	  solver_mode->insertItem("Thomas TDMA",SolverParams::ThomasTDMA);
	  solver_mode->setCurrentItem(SolverParams::GaussSeidel);
	  connect(solver_mode,SIGNAL(activated(int)),this,SLOT(solver_change(int)));
	  QToolTip::add(solver_mode,_solver_explanation);
	  connect(solver_mode,SIGNAL(activated(int)),
	    &areawin,SLOT(restart_solver()));
	
	start_button=new QPushButton("&Start",parent);
	  start_button->setAccel(ALT + Key_S);
	  connect(start_button,SIGNAL(clicked()),
		this,SLOT(start_clicked()));
      QToolTip::add(start_button,"Start / stop / continue calculation.");
	start_button_text=0;
	
	area_button=new QPushButton("&Area...",parent);
	  connect(area_button,SIGNAL(clicked()),&asetup,SLOT(map()));
	  QToolTip::add(area_button,"Specify area width and height.");
	
	eps_mant=new QSpinBox(0,99,1,parent);
	  eps_mant->setValue(1);
	  eps_mant->setMinimumWidth(35);
	  connect(eps_mant,SIGNAL(valueChanged(int)),
	    &areawin,SLOT(restart_solver()));
	eps_exp=new QSpinBox(-15,0,1,parent);
	  eps_exp->setValue(-6);
	  eps_exp->setMinimumWidth(40);
	  connect(eps_exp,SIGNAL(valueChanged(int)),
	    &areawin,SLOT(restart_solver()));
	
	shuffle_mode=new QComboBox(parent);
	  shuffle_mode->insertItem("[none]",SolverParams::NoShuffle);
	  shuffle_mode->insertItem("X+X-Y+Y-",SolverParams::XpXmYpYm);
	  shuffle_mode->insertItem("X+Y+X-Y-",SolverParams::XpYpXmYm);
	
	QHBoxLayout *solver_eps_layout=new QHBoxLayout();
	  solver_eps_layout->addWidget(eps_mant);
	  solver_eps_layout->addWidget(new QLabel("E",parent));
	  solver_eps_layout->addWidget(eps_exp);
	
	QHBoxLayout *solver_solver_layout=new QHBoxLayout();
	  solver_solver_layout->addWidget(wtmp=new QLabel(solver_mode,
	    "Sol&ver:",parent));
	  solver_solver_layout->addWidget(solver_mode);
	
	QGridLayout *solver_g_layout=new QGridLayout(4,2);
	  solver_g_layout->addWidget(limit_timeslice,0,1);
	  solver_g_layout->addWidget(wtmp=new QLabel(limit_timeslice,
		"&Timeslice:",parent),0,0);
	  QToolTip::add(wtmp,"Limit PDE solver time slice (msec)\n"
	    "(interrupted after whole iterations).");
	  solver_g_layout->addLayout(solver_eps_layout,1,1);
	  solver_g_layout->addWidget(wtmp=new QLabel(eps_exp,
		"&Epsilon:",parent),1,0);
	  QToolTip::add(wtmp,"PDE solver convergence threshold.");
	  solver_g_layout->addWidget(relax_omega,2,1);
	  solver_g_layout->addWidget(wtmp=new QLabel(relax_omega,
		"&Omega:",parent),2,0);
	  QToolTip::add(wtmp,"Over-relaxation factor.");
	  solver_g_layout->addWidget(shuffle_mode,3,1);
	  solver_g_layout->addWidget(wtmp=new QLabel(shuffle_mode,
		"&Shuffle:",parent),3,0);
	  QToolTip::add(wtmp,"Shuffle: change sweep direction\n"
	    "during iterations.");
	  solver_g_layout->addWidget(limit_iterations,4,1);
	  solver_g_layout->addWidget(wtmp=new QLabel(limit_iterations,
		"&MaxIter:",parent),4,0);
	  QToolTip::add(wtmp,"Limit number of PDE\nsolver iterations.");
	
	periodic_bc=new QCheckBox("&Periodic BC",parent);
	  QToolTip::add(periodic_bc,
	    "Enable periodic boundary conditions.");
	  connect(periodic_bc,SIGNAL(toggled(bool)),
	    &areawin,SLOT(restart_solver()));
	
	QHBoxLayout *solver_bt_layout=new QHBoxLayout();
	  solver_bt_layout->addWidget(start_button);
	  solver_bt_layout->addWidget(area_button);
	
	QVBoxLayout *solver_layout=new QVBoxLayout(parent,margin,spacing);
	  solver_layout->addSpacing(15);
	  solver_layout->addLayout(solver_solver_layout);
	  solver_layout->addLayout(solver_g_layout);
	  solver_layout->addWidget(periodic_bc);
	  solver_layout->addLayout(solver_bt_layout);
	
	/******** AREA ********/
	QGroupBox *area_group=new QGroupBox("Area",this);
	parent=area_group;
	
	reset_button=new QPushButton("&Reset",parent);
	  connect(reset_button,SIGNAL(clicked()),
		this,SLOT(reset_clicked()));
      QToolTip::add(reset_button,
	    "Clear area.");
	zero_button=new QPushButton("&Zero",parent);
	  connect(zero_button,SIGNAL(clicked()),
		this,SLOT(zero_clicked()));
      QToolTip::add(zero_button,
	    "Zero all floating elements.");
	
	ncolors_sel=new QSpinBox(2,512,4,parent);
	  ncolors_sel->setValue(256);
	  connect(ncolors_sel,SIGNAL(valueChanged(int)),
	    &areawin,SLOT(change_ncolors(int)));
	
	colgrad=new ColorGradientDisplay(&areawin,parent);
	
	QHBoxLayout *area_bt_layout=new QHBoxLayout();
	  area_bt_layout->addWidget(reset_button);
	  area_bt_layout->addWidget(zero_button);
	
	QHBoxLayout *area_ncol_layout=new QHBoxLayout();
	  area_ncol_layout->addWidget(wtmp=new QLabel(
	    ncolors_sel,"N&Colors:",parent));
	  QToolTip::add(wtmp,"Set number of colors.");
	  area_ncol_layout->addWidget(ncolors_sel);
	
	QVBoxLayout *area_layout=new QVBoxLayout(parent,margin,spacing);
	  area_layout->addSpacing(15);
	  area_layout->addLayout(area_bt_layout);
	  area_layout->addLayout(area_ncol_layout);
	  area_layout->addWidget(colgrad);
	
	/******** DRAW ********/
	QGroupBox *draw_group=new QGroupBox("Draw",this);
	parent=draw_group;
	
	shape_select=new QComboBox(parent);
	  shape_select->insertItem("Point",DrawParams::Point);
	  shape_select->insertItem("Rectangle",DrawParams::Rectangle);
	  shape_select->insertItem("Ellipse",DrawParams::Ellipse);
	  QToolTip::add(shape_select,"Select shape to draw.");
	
	type_select=new QComboBox(parent);
	  QToolTip::add(type_select,
	    "Area element type.");
	  type_select->insertItem("Floating",EElement::Floating);
	  type_select->insertItem("Fixed",EElement::Fixed);
	  type_select->setCurrentItem(EElement::Fixed);
	
	potential_val=new QSpinBox(-1000000,1000000,10,parent);
	  potential_val->setValue(0);
	
	QGridLayout *draw_i_layout=new QGridLayout(3,2);
	  draw_i_layout->addWidget(wtmp=new QLabel(shape_select,
	    "Shape:",parent),0,0);
	  draw_i_layout->addWidget(shape_select,0,1);
	  draw_i_layout->addWidget(wtmp=new QLabel(type_select,
	    "Type:",parent),1,0);
	  draw_i_layout->addWidget(type_select,1,1);
	  draw_i_layout->addWidget(wtmp=new QLabel(potential_val,
		"Potential:",parent),2,0);
	  QToolTip::add(wtmp,"Enter potential value.");
	  draw_i_layout->addWidget(potential_val,2,1);
	
	QVBoxLayout *draw_layout=new QVBoxLayout(parent,margin,spacing);
	  draw_layout->addSpacing(15);
	  draw_layout->addLayout(draw_i_layout);
	
	/******** STATUS ********/
	
	//swin=new StatusWindow(this);
	//areawin.SetStatWindow(swin);
	
	/******** TOP LAYOUT ********/
	QVBoxLayout *toplayout=new QVBoxLayout(this,/*margin=*/3,/*spacing=*/0);
	  toplayout->addLayout(head_layout);
	  toplayout->addSpacing(2);
	  toplayout->addWidget(solver_group);
	  toplayout->addWidget(area_group);
	  toplayout->addWidget(draw_group);
	  //toplayout->addWidget(swin);
	  //toplayout->addSpacing(5);
	  //toplayout->addWidget(bpropd);
	
	startTimer(0);
}


MainWindow::~MainWindow()
{
	//delete swin;
}
